/*		Escape and unescape undesireable characters		HTEscap.c
**		================================

**	(c) COPYRIGHT MIT 1995. 
**	All Rights Reserved. http://www.w3.org/Consortium/Legal/" 
**
**  THIS SOFTWARE AND DOCUMENTATION IS PROVIDED "AS IS," AND COPYRIGHT 
**  HOLDERS MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED, 
**  INCLUDING BUT NOT LIMITED TO, WARRANTIES OF MERCHANTABILITY OR FITNESS 
**  FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE OR 
**  DOCUMENTATION WILL NOT INFRINGE ANY THIRD PARTY PATENTS, COPYRIGHTS, 
**  TRADEMARKS OR OTHER RIGHTS. 
**
**  COPYRIGHT HOLDERS WILL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL
**  OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE OR 
**  DOCUMENTATION. 
**
** history:
**	This module extracts the escape and unescape functions from the
**  cern webserver module HTParse.c. The code has been modified slightly
**  to remove constructs that prevented a clean compile in the Bull SRM
**  environment.
** 
*/

#include "HTEscape.h"

/*      Escape undesirable characters using %       HTEscape()
**      -------------------------------------
**
**  This function takes a pointer to a string in which
**  some characters may be unacceptable unescaped.
**  It returns a string which has these characters
**  represented by a '%' character followed by two hex digits.
**
**  In the tradition of being conservative in what you do and liberal
**  in what you accept, we encode some characters which in fact are
**  allowed in URLs unencoded -- so DON'T use the table below for
**  parsing! 
**
**  Unlike HTUnEscape(), this routine returns a malloced string.
**
*/

#define HEX_ESCAPE '%'
#ifndef TOASCII
#define TOASCII(c) (c)
#define FROMASCII(c) (c)
#endif
 
/* Not BOTH static AND const at the same time in gcc :-(, Henrik 18/03-94 
**  code gen error in gcc when making random access to
**  static const table(!!)  */
/* PRIVATE CONST unsigned char isAcceptable[96] = */
static unsigned char isAcceptable[96] =

/* Overencodes */
/*  Bit 0       xalpha      -- see HTFile.h
**  Bit 1       xpalpha     -- as xalpha but with plus.
**  Bit 2 ...   path        -- as xpalpha but with /
*/
    /*   0 1 2 3 4 5 6 7 8 9 A B C D E F */
    {    0,0,0,0,0,0,0,0,0,0,7,6,0,7,7,4,   /* 2x   !"#$%&'()*+,-./  */
         7,7,7,7,7,7,7,7,7,7,0,0,0,0,0,0,   /* 3x  0123456789:;<=>?  */
         7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,   /* 4x  @ABCDEFGHIJKLMNO  */
         7,7,7,7,7,7,7,7,7,7,7,0,0,0,0,7,   /* 5X  PQRSTUVWXYZ[\]^_  */
         0,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,   /* 6x  `abcdefghijklmno  */
         7,7,7,7,7,7,7,7,7,7,7,0,0,0,0,0 }; /* 7X  pqrstuvwxyz{\}~  DEL */

static char *hex = "0123456789ABCDEF";
static char escaped[250];

extern char * HTEscape( char *str, unsigned char mask)
{
#define ACCEPTABLE(a)   ( a>=32 && a<128 && ((isAcceptable[a-32]) & mask))
    char * p;
    char * q;
    char * result = escaped;
    int unacceptable = 0;
    for(p=str; *p; p++)
        if (!ACCEPTABLE((unsigned char)TOASCII(*p)))
        unacceptable++;
/*    result = (char *) malloc(p-str + unacceptable+ unacceptable + 1); */
/*    if (result == NULL) outofmem(__FILE__, "HTEscape"); */
    for(q=result, p=str; *p; p++) {
        unsigned char a = TOASCII(*p);
    if (!ACCEPTABLE(a)) {
        *q++ = HEX_ESCAPE;  /* Means hex commming */
        *q++ = hex[a >> 4];
        *q++ = hex[a & 15];
    }
    else *q++ = *p;
    }
    *q++ = 0;           /* Terminate */
    return result;
}


/*      Decode %xx escaped characters           HTUnEscape()
**      -----------------------------
**
**  This function takes a pointer to a string in which some
**  characters may have been encoded in %xy form, where xy is
**  the acsii hex code for character 16x+y.
**  The string is converted in place, as it will never grow.
*/

static char from_hex(char c)
{
    return  c >= '0' && c <= '9' ?  c - '0' 
            : c >= 'A' && c <= 'F'? c - 'A' + 10
            : c - 'a' + 10; /* accept small letters just in case */
}

extern char * HTUnEscape( char *str)
{
    char * p = str;
    char * q = str;

    if (!str) {                       /* Just for safety ;-) */
/*    if (URI_TRACE)													*/
/*        fprintf(stderr, "HTUnEscape.. Called with NULL argument.\n"); */
    return "";
    }
    while(*p) {
        if (*p == HEX_ESCAPE) {
        p++;
        if (*p) *q = from_hex(*p++) * 16;
        if (*p) *q = FROMASCII(*q + from_hex(*p++));
        q++;
    } else {
        *q++ = *p++; 
    }
    }
    
    *q++ = 0;
    return str;
    
} /* HTUnEscape */

